__title__ = 'lightwood'
__package_name__ = 'lightwood'
__version__ = '24.3.3.1'
__description__ = "Lightwood is a toolkit for automatic machine learning model building"
__email__ = "community@mindsdb.com"
__author__ = 'MindsDB Inc'
__github__ = 'https://github.com/mindsdb/lightwood'
__pypi__ = 'https://pypi.org/project/lightwood'
__license__ = 'GPL-3.0'
__copyright__ = 'Copyright 2019- mindsdb'
